"""Baseclasses for services using asyncio as eventloop"""
from typing import Union, MutableMapping, Any, Coroutine, Optional
from pathlib import Path
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import tempfile
import logging
import functools

import signal as posixsignal
import asyncio
import toml

from datastreamcorelib.abstract import ZMQSocketType, ZMQSocketDescription
from datastreamcorelib.utils import create_heartbeat_message


from .zmqwrappers import PubSubManager, pubsubmanager_factory
from .compat import cast_task

LOGGER = logging.getLogger(__name__)
EXCEPTION_EXITCODE = 200
SIGNALS_HOOKED = False


# See https://github.com/python/mypy/issues/5374 why the typing ignore
@dataclass  # type: ignore
class BaseService(ABC):
    """Baseclass for services using asyncio as eventloop"""

    disable_signals_hook: bool = field(init=False, default=False)
    _exitcode: Union[None, int] = field(init=False, default=None)
    _quitevent: asyncio.Event = field(init=False, default_factory=asyncio.Event, repr=False)

    @abstractmethod
    async def setup(self) -> None:
        """Called once by the run method before wating for exitcode to be set"""
        raise NotImplementedError()

    @abstractmethod
    async def teardown(self) -> None:
        """Called once by the run method before exiting"""
        raise NotImplementedError()

    @abstractmethod
    def reload(self) -> None:
        """Called whenever the service is told to reload (usually via SIGHUP), setup might want to call this"""
        raise NotImplementedError()

    def quit(self, exitcode: int = 0) -> None:
        """set the exitcode which as side-effect will tell the run method to do teardown and finally exit"""
        LOGGER.debug("called with code={}".format(exitcode))
        self._exitcode = exitcode
        self._quitevent.set()

    def hook_signals(self) -> None:
        """Hook handlers for signals and the default exception handler"""
        global SIGNALS_HOOKED  # pylint: disable=W0603
        if self.disable_signals_hook:
            LOGGER.info("Signal hooking disabled by property")
            return
        if SIGNALS_HOOKED:
            LOGGER.warning("Already hooked")
            return
        loop = asyncio.get_event_loop()
        loop.set_exception_handler(self.default_exception_handler)
        loop.add_signal_handler(posixsignal.SIGINT, self.quit)
        loop.add_signal_handler(posixsignal.SIGTERM, self.quit)
        try:
            loop.add_signal_handler(posixsignal.SIGHUP, self.reload)
        except AttributeError:
            # Windows does not implement all signals
            pass
        SIGNALS_HOOKED = True

    def default_exception_handler(self, loop: asyncio.AbstractEventLoop, context: MutableMapping[str, Any]) -> None:
        """Default exception handler, logs the exception and quits (forcefully)"""
        msg = context.get("exception", context["message"])
        if isinstance(msg, Exception):
            LOGGER.exception("Unhandled in loop {}".format(loop), exc_info=msg)
        else:
            LOGGER.error("Called with context message {} in loop {}".format(msg, loop))
        # Make sure we get killed if quit hangs...
        BaseService.set_exit_alarm()
        # Call quit with the defined exit code
        self.quit(EXCEPTION_EXITCODE)

    @classmethod
    def set_exit_alarm(cls, timeout: int = 2) -> bool:
        """Clears handlers from alarm and sets an alarm signal for hard exit on timeout

        returns False if alarm could not be set. Yes, it can only take full integer seconds"""
        try:
            # Try to make sure alarm is not handled and set alarm to make sure teardown doesn't hang
            loop = asyncio.get_event_loop()
            loop.remove_signal_handler(posixsignal.SIGALRM)
            posixsignal.signal(posixsignal.SIGALRM, posixsignal.SIG_DFL)
            posixsignal.alarm(timeout)
            return True
        except AttributeError:
            # Windows does not implement all signals
            return False

    @classmethod
    def clear_exit_alarm(cls) -> bool:
        """Clear pending alarm"""
        try:
            posixsignal.signal(posixsignal.SIGALRM, posixsignal.SIG_DFL)
            posixsignal.alarm(0)
            return True
        except AttributeError:
            # Windows does not implement all signals
            return False

    async def run(self) -> int:
        """Main entrypoint, should be called with asyncio.get_event_loop().run_until_complete()"""
        await self.setup()
        await self._quitevent.wait()
        if self._exitcode is None:
            LOGGER.error("Got quitevent but exitcode is not set")
            self._exitcode = 1
        # Try to make sure teardown does not hang
        BaseService.set_exit_alarm()
        await self.teardown()
        return self._exitcode


@dataclass
class SimpleServiceMixin:
    """Mixin for a bit of automagics for heartbeats, config loading etc"""

    configpath: Path
    psmgr: PubSubManager = field(init=False, default_factory=pubsubmanager_factory)
    config: MutableMapping[str, Any] = field(init=False, default_factory=dict)
    _tasks: MutableMapping[str, "asyncio.Task[Any]"] = field(init=False, default_factory=dict)

    def create_task(self, coro: Coroutine[Any, Any, Any], *, name: Optional[str] = None) -> "asyncio.Task[Any]":
        """Helper to wrap asyncios create_task so that we always handle the exception and track long-running tasks"""
        task = asyncio.get_event_loop().create_task(coro)
        if name:
            if name in self._tasks:
                raise ValueError(f"name {name} is already tracked")
        else:
            name = f"Task-{id(task):02x}"

        task = cast_task(task)
        try:
            task.set_name(name)
        except AttributeError:
            # Ignore the missing method in py36 and add the get_name wrapper
            task.get_name = functools.partial(lambda retname: retname, name)  # type: ignore

        def report_error_remove_tracking(task: "asyncio.Task[Any]") -> None:
            """done callback to bubble up errors and remove tracking"""
            # Remove from tracking (we monkeypatched get_name above)
            task = cast_task(task)
            name = task.get_name()
            del self._tasks[name]
            # Bubble up any exceptions
            try:
                exc = task.exception()
                if exc:
                    LOGGER.error("Task {} raised exception {}".format(task, exc))
                    raise exc
            except asyncio.CancelledError:
                LOGGER.error("Task {} did not handle cancellation".format(task))

        self._tasks[name] = task
        task.add_done_callback(report_error_remove_tracking)
        return task

    async def _heartbeat_task(self) -> None:
        """Send a periodic heartbeat"""
        try:
            while self.psmgr.default_pub_socket and not self.psmgr.default_pub_socket.closed:
                await self.psmgr.publish_async(create_heartbeat_message())
                await asyncio.sleep(1)
            LOGGER.warning("Lost self.psmgr.default_pub_socket before heartbeat was cancelled")
        except asyncio.CancelledError:
            LOGGER.debug("Cancelled")
            return

    def _resolve_default_pub_socket(self) -> None:
        """Resolves the path for default PUB socket and sets it to PubSubManager"""
        pub_default = "ipc://" + str(Path(tempfile.gettempdir()) / self.configpath.name.replace(".toml", "_pub.sock"))
        if "zmq" in self.config and "pub_sockets" in self.config["zmq"]:
            pub_default = self.config["zmq"]["pub_sockets"]
        sdesc = ZMQSocketDescription(pub_default, ZMQSocketType.PUB)
        sock = self.psmgr.sockethandler.get_socket(sdesc)
        LOGGER.debug("Setting psmgr@{} default pub socket to {} (sdesc={})".format(hex(id(self.psmgr)), sock, sdesc))
        self.psmgr.default_pub_socket = sock

    async def _stop_hbtask_graceful(self) -> None:
        """Stops the hb task if it's active"""
        if "HEARTBEAT" not in self._tasks:
            return
        await self.stop_named_task_graceful("HEARTBEAT")

    async def _restart_hb_task(self) -> None:
        """Stop and recreate hb task"""
        await self._stop_hbtask_graceful()
        self.create_task(self._heartbeat_task(), name="HEARTBEAT")

    def reload(self) -> None:
        """Load configs, restart sockets"""
        self.psmgr.sockethandler.close_all_sockets()
        with self.configpath.open("rt", encoding="utf-8") as filepntr:
            self.config = toml.load(filepntr)
        self._resolve_default_pub_socket()
        self.create_task(self._restart_hb_task())

    async def stop_named_task_graceful(self, taskname: str) -> Optional[Any]:
        """cancel the named task if it is running and return the result"""
        if taskname not in self._tasks:
            LOGGER.warning("task {} not found".format(taskname), stack_info=True)
            return None
        task = self._tasks[taskname]
        return await self.stop_task_graceful(task)

    async def stop_task_graceful(self, task: "asyncio.Task[Any]") -> Any:
        """cancel the given task if it is running and return the result"""
        if not task.done():
            LOGGER.info("Cancelling task {}".format(task))
            task.cancel()
        try:
            return await task
        except asyncio.CancelledError:
            LOGGER.error("Task {} did not handle cancellation".format(task))

    async def stop_lingering_tasks(self) -> None:
        """Stop all still lingering tasks and fetch their results"""
        # we modify the dictionary during iteration, make a copy of the values
        for task in list(self._tasks.values()):
            try:
                await self.stop_task_graceful(task)
                LOGGER.info("Task {} stopped".format(task))
            except Exception:  # pylint: disable=W0703
                LOGGER.exception("Task {} returned exception".format(task))

    async def teardown(self) -> None:
        """Stop all tasks, close all sockets"""
        await self._stop_hbtask_graceful()
        await self.stop_lingering_tasks()
        self.psmgr.sockethandler.close_all_sockets()


@dataclass
class SimpleService(SimpleServiceMixin, BaseService):
    """Simple service does a bit of automagics in setup"""

    async def setup(self) -> None:
        """Called once by run, just calls reload which loads our config"""
        self.hook_signals()
        self.reload()



"""Console scripts"""
import asyncio
import logging
import sys
import tempfile
import uuid
from pathlib import Path
from dataclasses import dataclass, field
from typing import Dict, List, Any, Union

import click
import toml


from datastreamcorelib.logging import init_logging
from datastreamcorelib.binpackers import ensure_utf8
from datastreamcorelib.pubsub import PubSubMessage, Subscription
from datastreamcorelib.datamessage import PubSubDataMessage
from datastreamcorelib.imagemessage import PubSubImageMessage
from datastreamservicelib.service import SimpleService


LOGGER = logging.getLogger(__name__)


@dataclass
class Publisher(SimpleService):
    """Publisher service"""

    topic: bytes
    send_count: int = field(default=-1)
    images: bool = field(default=False)

    def reload(self) -> None:
        """Create task for sending messages"""
        super().reload()
        self.create_task(self.message_sender(), name="sender")

    async def message_sender(self) -> None:
        """Send messages"""
        msgno = 0
        try:
            while self.psmgr.default_pub_socket and not self.psmgr.default_pub_socket.closed:
                msgno += 1
                msg: Union[None, PubSubImageMessage, PubSubDataMessage] = None
                if self.images:
                    msg = PubSubImageMessage(topic=self.topic)
                    msg.imginfo["format"] = "bgr8"
                    msg.imginfo["w"] = 1
                    msg.imginfo["h"] = 1
                    msg.imgdata = bytes((0, 0, 255))
                else:
                    msg = PubSubDataMessage(topic=self.topic)
                if not msg:
                    raise RuntimeError("We should not reach this ever")
                msg.data["msgno"] = msgno
                LOGGER.debug("Publishing {}".format(msg))
                await self.psmgr.publish_async(msg)
                if self.send_count > 0 and msgno > self.send_count:
                    return self.quit()
                await asyncio.sleep(0.1)
        except asyncio.CancelledError:
            pass
        return None


@dataclass
class CollectingSubscriber(SimpleService):
    """Subscriber service that track received messages"""

    topic: bytes
    recv_count: int = field(default=-1)
    messages_by_sub: Dict[uuid.UUID, List[PubSubMessage]] = field(default_factory=dict)
    decoder_class: Any = field(default=PubSubDataMessage)

    def reload(self) -> None:
        """Create subscription"""
        super().reload()
        sub = Subscription(
            self.config["zmq"]["sub_sockets"],
            self.topic,
            self.success_callback,
            decoder_class=self.decoder_class,
            isasync=True,
        )
        self.messages_by_sub[sub.trackingid] = []
        self.psmgr.subscribe_async(sub)

    async def success_callback(self, sub: Subscription, msg: PubSubMessage) -> None:
        """Append the message to the correct list"""
        if sub.trackingid not in self.messages_by_sub:
            self.messages_by_sub[sub.trackingid] = []
        self.messages_by_sub[sub.trackingid].append(msg)
        LOGGER.debug("Got msg {}".format(msg))
        if self.recv_count > 0 and len(self.messages_by_sub[sub.trackingid]) >= self.recv_count:
            self.quit()

    async def teardown(self) -> None:
        """Clean up the messages_by_sub"""
        self.messages_by_sub = {}
        return await super().teardown()


@click.command()
@click.option("-s", "--socket_uri", help="For example ipc:///tmp/publisher.sock", required=True)
@click.option("-t", "--topic", help="The topic to use for sending", required=True)
@click.option("-i", "--images", help="Send images (one red pixel)", is_flag=True)
@click.option("-c", "--count", help="Number of messages to send", type=int, default=-1)
def publisher_cli(socket_uri: str, topic: str, count: int, images: bool) -> None:
    """CLI entrypoint for publisher"""
    init_logging(logging.DEBUG)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.debug("sys.argv={}".format(sys.argv))
    with tempfile.TemporaryDirectory() as tmpdir:
        configpath = Path(tmpdir) / "config.toml"
        LOGGER.debug("writing file {}".format(configpath))
        with open(configpath, "wt", encoding="utf-8") as fpntr:
            toml.dump({"zmq": {"pub_sockets": [socket_uri]}}, fpntr)
        pub_instance = Publisher(configpath, ensure_utf8(topic), count, images)
        exitcode = asyncio.get_event_loop().run_until_complete(pub_instance.run())
    sys.exit(exitcode)


@click.command()
@click.option("-s", "--socket_uri", help="Must be same the publisher uses", required=True)
@click.option("-t", "--topic", help="The topic to use for receiving, must match topic used in publisher", required=True)
@click.option("-i", "--images", help="Use image decoder", is_flag=True)
@click.option("-c", "--count", help="Number of messages to send", type=int, default=-1)
def subscriber_cli(socket_uri: str, topic: str, count: int, images: bool) -> None:
    """CLI entrypoint for subscriber"""
    init_logging(logging.DEBUG)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.debug("sys.argv={}".format(sys.argv))
    with tempfile.TemporaryDirectory() as tmpdir:
        configpath = Path(tmpdir) / "config.toml"
        LOGGER.debug("writing file {}".format(configpath))
        with open(configpath, "wt", encoding="utf-8") as fpntr:
            toml.dump({"zmq": {"sub_sockets": [socket_uri]}}, fpntr)
        sub_instance = CollectingSubscriber(configpath, ensure_utf8(topic), count)
        if images:
            sub_instance.decoder_class = PubSubImageMessage
        exitcode = asyncio.get_event_loop().run_until_complete(sub_instance.run())
    sys.exit(exitcode)



"""REQuest/REPly helpers"""
from typing import Union, Any, cast
import asyncio
import logging
import tempfile
from pathlib import Path
from dataclasses import dataclass


import zmq  # type: ignore  # Get rid of error: Cannot find implementation or library stub for module
from datastreamcorelib.abstract import ZMQSocketDescription, ZMQSocketType, ZMQSocketUrisInputTypes, ZMQSocket
from datastreamcorelib.datamessage import PubSubDataMessage
from datastreamcorelib.reqrep import REQMixinBase, REPMixinBase

from .zmqwrappers import Socket
from .service import SimpleServiceMixin, SimpleService


LOGGER = logging.getLogger(__name__)
REQ_SEND_TIMEOUT = 0.5
REQ_RECV_TIMEOUT = 1.0
REP_SEND_TIMEOUT = 0.5


class REQMixin(SimpleServiceMixin, REQMixinBase):
    """Mixin for making REQuests"""

    def _get_request_socket(self, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes]) -> ZMQSocket:
        """Get the socket"""
        if isinstance(sockdef, ZMQSocket):
            return sockdef
        sdesc = ZMQSocketDescription(sockdef, ZMQSocketType.REQ)
        return self.psmgr.sockethandler.get_socket(sdesc)

    def _do_reqrep_blocking(
        self, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes], msg: PubSubDataMessage
    ) -> PubSubDataMessage:
        raise TypeError("Not supported")

    async def _do_reqrep_async(
        self, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes], msg: PubSubDataMessage
    ) -> PubSubDataMessage:
        """Do the actual REQuest and get the REPly (async context)"""
        sock = cast(Socket, self._get_request_socket(sockdef))
        try:
            await asyncio.wait_for(sock.send_multipart(msg.zmq_encode()), timeout=REQ_SEND_TIMEOUT)
            resp_parts = await asyncio.wait_for(sock.recv_multipart(), timeout=REQ_RECV_TIMEOUT)
            return PubSubDataMessage.zmq_decode(resp_parts)
        except (asyncio.TimeoutError, zmq.ZMQBaseError):
            sock.close()
            raise

    async def send_command(
        self, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes], cmd: str, *args: Any, raise_on_insane: bool = False
    ) -> PubSubDataMessage:
        """shorthand for send_command_async"""
        return await self.send_command_async(sockdef, cmd, *args, raise_on_insane=raise_on_insane)


class REPMixin(SimpleServiceMixin, REPMixinBase):
    """Mixin for making REPlies"""

    def _resolve_default_rep_socket_uri(self) -> str:
        """Resolves the path for default PUB socket and sets it to PubSubManager"""
        pub_default = "ipc://" + str(Path(tempfile.gettempdir()) / self.configpath.name.replace(".toml", "_rep.sock"))
        if "zmq" in self.config and "rep_sockets" in self.config["zmq"]:
            pub_default = self.config["zmq"]["rep_sockets"]
        return pub_default

    async def _reply_task(self, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes, None]) -> None:
        """Bind the given socket and start dealing with REQuests"""
        sdesc = None
        if isinstance(sockdef, ZMQSocket):
            sock = cast(Socket, sockdef)
        else:
            if sockdef is None:
                sockdef = self._resolve_default_rep_socket_uri()
            sdesc = ZMQSocketDescription(sockdef, ZMQSocketType.REP)
            sock = cast(Socket, self.psmgr.sockethandler.get_socket(sdesc))
        try:
            while not sock.closed:
                msgparts = await sock.recv_multipart()
                replymsg = await self.handle_rep_async(msgparts, sdesc)
                await asyncio.wait_for(sock.send_multipart(replymsg.zmq_encode()), timeout=REP_SEND_TIMEOUT)
            LOGGER.info("{} closed from under us".format(sock))
        except asyncio.CancelledError:
            LOGGER.debug("cancelled")
        finally:
            sock.close()

    def reload(self) -> None:
        """Load configs, restart sockets"""
        super().reload()
        # Create reply handler in the default uri
        self.create_task(self._reply_task(None), name="DEFAULT_REP")


@dataclass  # pylint: disable=R0901
class FullService(REPMixin, REQMixin, SimpleService):
    """Service with REQuest and REPly mixins applied"""



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['datastreamservicelib']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'datastreamcorelib>=1.2,<2.0',
 'pyzmq>=19.0,<20.0',
 'toml>=0.10,<0.11',
 'tomlkit>=0.6,<0.7']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['testpublisher = '
                     'datastreamservicelib.console:publisher_cli',
                     'testsubscriber = '
                     'datastreamservicelib.console:subscriber_cli']}

setup_kwargs = {
    'name': 'datastreamservicelib',
    'version': '1.10.1',
    'description': 'AsyncIO eventloop helpers and Abstract Base Classes for making services that use ZMQ nice, easy and DRY',
    'long_description': '====================\ndatastreamservicelib\n====================\n\nAsyncIO eventloop helpers and Abstract Base Classes for making services that use ZMQ nice, easy and DRY\n\nUsage\n-----\n\nUse the CookieCutter template at https://gitlab.com/advian-oss/python-datastreamserviceapp_template\n\nYou can also take a look at src/datastreamservicelib/console.py for some very simple test examples.\n\nDocker\n------\n\nFor more controlled deployments and to get rid of "works on my computer" -syndrome, we always\nmake sure our software works under docker.\n\nIt\'s also a quick way to get started with a standard development environment.\n\nSSH agent forwarding\n^^^^^^^^^^^^^^^^^^^^\n\nWe need buildkit_::\n\n    export DOCKER_BUILDKIT=1\n\n.. _buildkit: https://docs.docker.com/develop/develop-images/build_enhancements/\n\nAnd also the exact way for forwarding agent to running instance is different on OSX::\n\n    export DOCKER_SSHAGENT="-v /run/host-services/ssh-auth.sock:/run/host-services/ssh-auth.sock -e SSH_AUTH_SOCK=/run/host-services/ssh-auth.sock"\n\nand Linux::\n\n    export DOCKER_SSHAGENT="-v $SSH_AUTH_SOCK:$SSH_AUTH_SOCK -e SSH_AUTH_SOCK"\n\nCreating a development container\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nBuild image, create container and start it::\n\n    docker build --ssh default --target devel_shell -t datastreamservicelib:devel_shell .\n    docker create --name datastreamservicelib_devel -v `pwd`":/app" -it -v /tmp:/tmp `echo $DOCKER_SSHAGENT` datastreamservicelib:devel_shell\n    docker start -i datastreamservicelib_devel\n\npre-commit considerations\n^^^^^^^^^^^^^^^^^^^^^^^^^\n\nIf working in Docker instead of native env you need to run the pre-commit checks in docker too::\n\n    docker exec -i datastreamservicelib_devel /bin/bash -c "pre-commit install"\n    docker exec -i datastreamservicelib_devel /bin/bash -c "pre-commit run --all-files"\n\nYou need to have the container running, see above. Or alternatively use the docker run syntax but using\nthe running container is faster::\n\n    docker run -it --rm -v `pwd`":/app" datastreamservicelib:devel_shell -c "pre-commit run --all-files"\n\nTest suite\n^^^^^^^^^^\n\nYou can use the devel shell to run py.test when doing development, for CI use\nthe "tox" target in the Dockerfile::\n\n    docker build --ssh default --target tox -t datastreamservicelib:tox .\n    docker run -it --rm -v `pwd`":/app" `echo $DOCKER_SSHAGENT` datastreamservicelib:tox\n\nProduction docker\n^^^^^^^^^^^^^^^^^\n\nThere\'s a "production" target as well for quick running of testsubscriber and testpublisher\n\n    docker build --ssh default --target production -t datastreamservicelib:latest .\n    docker run --rm -it -v /tmp:/tmp datastreamservicelib:latest testpublisher -s ipc:///tmp/test_pub.sock -t foo\n    docker run --rm -it -v /tmp:/tmp datastreamservicelib:latest testsubscriber -s ipc:///tmp/test_pub.sock -t foo\n\nNote that on non-linux platforms the IPC sockets may not as expected between host and container over volume mounts.\n\n\nLocal Development\n-----------------\n\nTLDR:\n\n- Create and activate a Python 3.6 virtualenv (assuming virtualenvwrapper)::\n\n    mkvirtualenv -p `which python3.6` my_virtualenv\n\n- change to a branch::\n\n    git checkout -b my_branch\n\n- install Poetry: https://python-poetry.org/docs/#installation\n- Install project deps and pre-commit hooks::\n\n    poetry install\n    pre-commit install\n    pre-commit run --all-files\n\n- Ready to go, try the following::\n\n    testpublisher --help\n    testsubscriber --help\n\nUse Python 3.6 for development since it\'s the lowest supported version so you don\'t accidentally\nuse features only available in higher versions and then have to re-do everything when CI tests fail\non 3.6.\n\nRemember to activate your virtualenv whenever working on the repo, this is needed\nbecause pylint and mypy pre-commit hooks use the "system" python for now (because reasons).\n\nRunning "pre-commit run --all-files" and "py.test -v" regularly during development and\nespecially before committing will save you some headache.\n',
    'author': 'Eero af Heurlin',
    'author_email': 'eero.afheurlin@iki.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/advian-oss/python-datastreamservicelib/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Compatibility stuff for old python versions"""
from typing import TypeVar, cast, TYPE_CHECKING
import asyncio

T = TypeVar("T")  # pylint: disable=C0103
if TYPE_CHECKING:
    from typing import Any  # pylint: disable=C0412


class NamedTask(asyncio.Task):  # type: ignore
    """wrapper for mypy checking in older python versions"""

    def get_name(self) -> str:
        """dummy"""
        raise NotImplementedError()

    def set_name(self, __value: object) -> None:
        """dummy"""
        raise NotImplementedError()


def cast_task(task: "asyncio.Task[T]") -> "NamedTask[T]":  # type: ignore
    """Cast task to NamedTask"""
    return cast("NamedTask[Any]", task)  # type: ignore



"""Implementations for the Abstract Base Classes from datastreamcorelib"""
from typing import cast, Any, Iterable, Dict, Callable, Awaitable, Union  # pylint: disable=W0611
from dataclasses import dataclass, field
import asyncio
import logging

import zmq  # type: ignore  # Get rid of error: Cannot find implementation or library stub for module
import zmq.asyncio  # type: ignore  # Get rid of error: Cannot find implementation or library stub for module

from datastreamcorelib.abstract import ZMQSocket, ZMQSocketType, BaseSocketHandler, ZMQSocketDescription
from datastreamcorelib.pubsub import (
    Subscription,
    BasePubSubManager,
    PubSubMessage,
    NoDefaultSocketError,
)
from datastreamcorelib.abstract import ZMQSocketUrisInputTypes

LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True)
class Socket(ZMQSocket):
    """Wrapper for the raw AIO socket"""

    socket_type: ZMQSocketType
    _raw_socket: Any = field(default=None, init=False, repr=True)

    def __post_init__(self) -> None:
        """Create the RAW socket"""
        # Map our abstract types to the implementation type
        if self.socket_type == ZMQSocketType.PUB:
            raw_socket_type = zmq.PUB  # pylint: disable=E1101
        if self.socket_type == ZMQSocketType.SUB:
            raw_socket_type = zmq.SUB  # pylint: disable=E1101
        if self.socket_type == ZMQSocketType.REQ:
            raw_socket_type = zmq.REQ  # pylint: disable=E1101
        if self.socket_type == ZMQSocketType.REP:
            raw_socket_type = zmq.REP  # pylint: disable=E1101
        object.__setattr__(self, "_raw_socket", zmq.asyncio.Context.instance().socket(raw_socket_type))

    def subscribe(self, topic: bytes) -> None:
        """Subscribe to a topic"""
        self._raw_socket.subscribe(topic)
        LOGGER.debug("{} subscribed to {!r}".format(self, topic))

    @property
    def closed(self) -> bool:
        """Tells if the socket is open or closed"""
        return bool(self._raw_socket.closed)

    def close(self) -> None:
        """Close the socket"""
        self._raw_socket.close()

    def __getattr__(self, key: str) -> Any:
        """Pass all but our abstractions to the raw socket"""
        return getattr(self._raw_socket, key)

    def __dir__(self) -> Iterable[str]:
        """Expose the raw socket dir too"""
        retval = set(self._raw_socket.__dir__())
        retval.add("__post_init__")
        retval.add("socket_type")
        retval.add("_raw_socket")
        return list(retval)

    def __hash__(self) -> int:
        """Make a hash based on the base sockets one"""
        return int(self._raw_socket.__hash__()) + self.socket_type.value


@dataclass
class SocketHandler(BaseSocketHandler):
    """Implement socket fetching"""

    def get_socket(self, desc: ZMQSocketDescription) -> ZMQSocket:
        """Get socket by address, or bind new one."""
        sock = self._get_cached_socket(desc)
        if sock:
            if not sock.closed:
                return sock
            # Clear the cache and create a new socket.
            del self._sockets_by_desc[desc]

        sock = Socket(desc.sockettype)
        # Bind/connect as needed
        if desc.sockettype not in (ZMQSocketType.SUB, ZMQSocketType.PUB, ZMQSocketType.REQ, ZMQSocketType.REP):
            raise RuntimeError("Don't know how to handle type {}".format(desc.sockettype))
        if desc.sockettype in (ZMQSocketType.SUB, ZMQSocketType.REQ):
            for addr in desc.socketuris:
                sock.connect(addr)
                LOGGER.debug("connected {} to {!r}".format(sock, addr))
        if desc.sockettype in (ZMQSocketType.PUB, ZMQSocketType.REP):
            for addr in desc.socketuris:
                sock.bind(addr)
                LOGGER.debug("bound {} to {!r}".format(sock, addr))

        self._set_cached_socket(desc, sock)
        return sock


@dataclass
class PubSubManager(BasePubSubManager):
    """
    Manage subscriptions, sockets and coroutines to read from the sockets
    """

    _socket_reader_tasks: Dict[Socket, "asyncio.Task[Callable[[Socket], Awaitable[None]]]"] = field(
        default_factory=dict, init=False
    )

    def __post_init__(self) -> None:
        """Instantiate the sockethandler"""
        # We do *not* want to use the singleton for sockethandler, it messes things up for running multiple
        # services in same asyncio loop
        self.sockethandler = self.handlerklass()

    def subscribe(self, sub: Subscription) -> None:
        """Add the subscription to our handlers and make sure socket reader is active"""
        super().subscribe(sub)
        sdesc = ZMQSocketDescription(sub.socketuris, ZMQSocketType.SUB)
        sock = cast(Socket, self.sockethandler.get_socket(sdesc))

        def reader_task_done(task: "asyncio.Task[Any]") -> None:
            """done callback for the reader task"""
            # raise exception if there was any
            task.result()

        if sock not in self._socket_reader_tasks or self._socket_reader_tasks[sock].done():
            # Add the reader task to the mainloop
            self._socket_reader_tasks[sock] = cast(
                "asyncio.Task[Callable[[Socket], Awaitable[None]]]",
                asyncio.get_event_loop().create_task(self._socket_reader(sock)),
            )
            self._socket_reader_tasks[sock].add_done_callback(reader_task_done)

    def subscribe_async(self, sub: Subscription) -> None:
        """Wrapper for subscribe that forces the subscription.isasync to be true"""
        if not sub.isasync:
            object.__setattr__(sub, "isasync", True)
        return self.subscribe(sub)

    async def _socket_reader(self, sock: Socket) -> None:
        """Waits and dispatches messages as long as the socket is open"""
        try:
            while not sock.closed:
                # LOGGER.debug("waiting message on {}".format(sock))
                msgparts = await sock.recv_multipart()
                # LOGGER.debug("got message on {}".format(sock))
                await self.raw_zmq_callback_async(sock, msgparts)
            LOGGER.debug("socket closed")
        except asyncio.CancelledError:
            # Handle cancellations gracefully
            LOGGER.debug("cancelled")

    async def publish_async(
        self, msg: PubSubMessage, sockdef: Union[ZMQSocket, ZMQSocketUrisInputTypes, None] = None
    ) -> None:
        """Publish a message, optionally take a socket definition of where to publish"""
        if isinstance(sockdef, ZMQSocket):
            ret = await cast(Socket, sockdef).send_multipart(msg.zmq_encode())
            return cast(None, ret)

        if sockdef is None:
            if not self.default_pub_socket:
                raise NoDefaultSocketError()
            ret = await cast(Socket, self.default_pub_socket).send_multipart(msg.zmq_encode())
            return cast(None, ret)

        sdesc = ZMQSocketDescription(sockdef, ZMQSocketType.PUB)
        sock = cast(Socket, self.sockethandler.get_socket(sdesc))
        ret = await sock.send_multipart(msg.zmq_encode())
        return cast(None, ret)

    def _publish_actual(self, msg: PubSubMessage, sock: ZMQSocket) -> None:
        """Implementation for actually handling the send, creates task to the eventloop"""
        # Fire-and-forget publish
        LOGGER.warning("You should use `await psmgr.publish_async` instead of just publish", stack_info=True)
        task = asyncio.get_event_loop().create_task(self.publish_async(msg, sock))

        def send_task_done(task: "asyncio.Task[Any]") -> None:
            """done callback for the send task"""
            # raise exception if there was any
            try:
                task.result()
            except asyncio.CancelledError:
                # Handle cancellations gracefully
                LOGGER.warning("task {} was cancelled for some reason, it should not have been".format(task))

        task.add_done_callback(send_task_done)


def pubsubmanager_factory() -> PubSubManager:
    """Get the default asyncio compatible PubSubManager with proper cast, requiring no arguments"""
    # We do *not* want to use the singleton for PubSubManager, it messes things up for running multiple
    # services in same asyncio loop
    return PubSubManager(SocketHandler)
